function[s,posdef,k,nnzPC,nnzLU] = ppcg(grad,H,A,b,options,fdata,DM,DG,tolA);
%PPCG	Preconditioned conjugate gradients with linear equalities.
%
%	[s,posdef,k] = ppcg(grad,H,A,b,options,fdata,D,DG);;
%
% 	A projected preconditioned conjugate gradient procedure
%	applied to
%
%	           min{ q(s) = grad's + .5s'Ms: As = b} 
%
%	where  M = DM*H*DM + DG.
%
%	DM,DG are pos. diagonal matrices (sparse format)
%       Symmetric matrix H is either explicit (sparse) or
%       implicitly available through options parameter 'mtxmpy'
%
%	INPUT:
%	_____
%	grad	- The linear term in the quadratic q(p) - see above.
%	H	- Hessian matrix (can be virtual).
%       A	- constraint mtx
%	b	- rhs
%	options - Named parameter list
%       fdata   - Sparse matrix of information (possibly) reqd by mtxmpy.
%       DM      - Positive diagonal matrix (sparse format).
%       DG      - A non-negative diagonal matrix (sparse format).
%
%	OUTPUT:
%	_______
%	s	- The solution.
%	posdef	- posdef = 1 then only pos. curvature has been ``sighted''
%		  posdef = 0 the neg. curvature discovered - p
%		  is dirn of neg. curvature.
%	k	- The number of CG iterations. 
%
%	******************************************************************
%	*              Thomas F. Coleman, 7-1-96                         *
%	******************************************************************
%
% Initializations.
  if nargin < 3, error('ppcg requires at least 3 input arguments'); end
  [m,n] = size(A); 
  if m >= n, error(' ppcg requires m < n'); end
  if nargin < 4, b = zeros(m,1); end
  if b == [], b = zeros(m,1); end
  [mm,nn] = size(b);
  if mm ~= m, error('b is incompatible with A in function cgle2'); end
  if nargin < 5, options = []; end
  if nargin < 6, fdata = []; end
  if nargin < 7, DM = speye(n); end
  if nargin < 8, DG = sparse(1:n,1:n,zeros(n,1)); end
  if isempty(DM); DM = speye(n); end
  if isempty(DG); DG = sparse(1:n,1:n,zeros(n,1)); end
% if nargin < 7, R = []; end
% if nargin < 8, s = (1:n)'; end
% if p == [], p = (1:n)'; end
  pcmtx = lsotget(options,'pcmtx') ;
  pcf = lsotget(options,'pcflags') ;
  mtx = lsotget(options,'mtxmpy') ;
  tol = lsotget(options,'tol') ;
  kmax = lsotget(options,'itb') ;
  if pcmtx == [], pcmtx = 'preaug'; end
  if pcf == [], pcf = 0; end
  if nargin >= 9, pcf(2,1) = tolA; end
  if mtx == [], mtx = 'hmult'; end
  if tol == [], tol = 1e-3; end
  if kmax == [], kmax = n; end
%
% Get feasible
  y = zeros(n,1);
  if norm(b) > 1e-12
     y = feasibl(A,b);
     ww = DM*y; w = eval([mtx,'(ww,H,fdata)']);
     ww = DM*w +DG*y; grad = grad + ww;
   end
%
% Transform to get fundamental null basis
AT = A';
  pp = colmmd(AT); 
  [L,U,PT] = lu(AT(:,pp));
  A = A*PT'; grad = PT*grad;
%
% Project grad
  [g,LZ,UZ,pcolZ,PZ] = fztmult(A,grad);
  r = -g; nz = length(g); s = zeros(nz,1); val = 0;

% Compute preconditioner
  HH = [];
  if H ~= [];
     HH = DM*H*DM + DG; HH = PT*HH*PT';
   end
  [L,U,P,pcol] = feval(pcmtx,HH,pcf,A);
%
% Precondition .
  rr = [zeros(m,1);r];
  rhs = [rr;zeros(m,1)];
  zz = L\(P*rhs); zt(pcol,1) = U\zz; z = zt(m+1:n,1);
  znrm = norm(z);
  stoptol = tol*znrm;
  inner2 = 0; inner1 = r'*z; posdef = 1;
%
% PRIMARY LOOP.
  for k = 1:kmax
      if k==1
         d = z;
      else
         beta = inner1/inner2;
         d = z + beta*d;
      end
      dd = fzmult(A,d,LZ,UZ,pcolZ,PZ);
      dd = PT'*dd;
      ww = DM*dd; w = eval([mtx,'(ww,H,fdata)']);
      ww = DM*w +DG*dd; ww = PT*ww; w = fztmult(A,ww,LZ,UZ,pcolZ,PZ);
      denom = d'*w;
      if denom <= 0
         s = d/norm(d);
         s = y + PT'*fzmult(A,s,LZ,UZ,pcolZ,PZ);
         posdef = 0;
         return
      else
         alpha = inner1/denom;
         s = s+ alpha*d; r = r - alpha*w;
      end
%
%     Precondition
      rr =[zeros(m,1);r];  
      rhs = [rr;zeros(m,1)];
      zz = L\(P*rhs); zt(pcol,1) = U\zz; z = zt(m+1:n,1);
%
%     Exit?
      znorm = norm(z);
      if znorm <= stoptol,
         s = y +  PT'*fzmult(A,s,LZ,UZ,pcolZ,PZ);
         return;
      end
      inner2 = inner1; inner1 = r'*z;
  end
  s = y +  fzmult(A,s,LZ,UZ,pcolZ,PZ);
