  function[p,posdef,k] = pcg(R,DM,DG,g,kmax,tol);
%
% A PRECONDITIONED CONJUGATE GRADIENT ROUTINE
% TO APPROXIMATELY MINIMIZE THE QUADRATIC FUNCTION
% q(p) = .5p'Mp + g'p ( TERMINATING ACCORDING TO THE NASH/SOFER CRITERION
% WITH TOLERANCE tol). MATRIX M IS DEFINED,
%
%              M = DM*H*DM + DG
%
% WHERE DM,DG ARE NON-NEGATIVE DIAGONAL MATRICES (SPARSE FORMAT)
% AND SYMMETRIC MATRIX H IS
% AVAILABLE IMPLICITLY THROUGH THE
% USE OF SUBROUTINE Hmult (FOR MATRIX-VECTOR MULTIPLY). H IS
% SYMMETRIC BUT NOT NECESSARILY POSITIVE DEFINITE.
%
% IF A DIRECTION OF NON-POSITIVE CURVATURE IS REVEALED
% BEFORE THE STOPPING TOLERANCE IS SATISFIED, THEN
% pcg IS EXITED  WITH posdef = 0 AND p
% IS A DIRECTION OF NON-POSITIVE CURVATURE -
% USUALLY NEGATIVE CURVATURE, I.E. p'Hp < 0.   
% OTHERWISE, ONLY POSITIVE CURVATURE HAS BEEN DISCOVERED,
% THE NASH/SOFER STOPPING CRITERION HAS BEEN SATISFIED, AND posdef = 1.
%
% k IS THE NUMBER OF ITERATIONS (MATRIX MULTIPLIES) USED.
% 
% R IS AN UPPER TRIANGULAR (SPARSE)  NONSINGULAR MATRIX SUCH
% THAT R'R REPRESENTS A PRECONDITIONER OF H.  
% kmax IS THE MAXIMUM NUMBER OF ITERATION ALLOWED TO
% TRY AND SATISFY THE STOPPING CRITERIA.  
%
% THE MATRIX H IS IMPLICITLY AVAILABLE FOR MATRIX-VECTOR MULTIPLY
% THROUGH FUNCTION Hmult.
%
  n = length(g); r = -g; p = zeros(n,1); val = 0;
  w = R'\r; z = R\w;
  inner2 = 0; inner1 = r'*z;
  posdef = 1;
%
% PRIMARY LOOP.
  for k = 1:kmax
      if k==1
         d = z;
      else
         beta = inner1/inner2;
         d = z + beta*d;
      end
      ww = DM*d; w = Hmult(ww); ww = DM*w +DG*d;
      denom = d'*ww;
      if denom <= 0
         p = d/norm(d);
         posdef = 0;
         break
      else
         alpha = inner1/denom;
         p = p+ alpha*d;
         r = r - alpha*ww;
      end
      valp = val;
      val = .5*(p'*r + p'*g);
      if abs(val - valp) <= tol*abs(val)
         break
      end
      w = R'\r; z = R\w; 
      inner2 = inner1;
      inner1 = r'*z;
  end
%
% LAST LINE OF pcg
