function [call,put,Svec,Tvec,sigval] = blspricefdsig(SO,X,r,T,sig,q,M,N)
%
%BLSPRICEFDSIG Black-Scholes put and call pricing for European Options using 
%	the Crank-Nicholson finite difference solution of Black-Scholes
%	Partial differential equation. SIG is allowed to be a function
%	of stock price SO, and/or time T.
%	
%	It returns different sort of output from BLSPRICE, here the
%	call and put values returned are in (SO x T) domain. SO, T
%	are not required to be input as matrices for getting a surface out.
%
%	Svec, Tvec are the stock and time space over which call and put
%	values are returned -- these can be used to mesh of plot the option value data.
%	The sigma value SIGVAL is also returned.
%
%	INPUT:
%	Note : X, r, q must be scalars!
%
%
%	If SO, T are vectors they should be equally spaced -- dS and dT will
%	be chosen as the distance between consecutive points.
%	SO(1) should be zero, and SO(N) should be "large" -- say 2*K or more.
%	No restriction on T is imposed. For accurate answers, both SO and
%	T should have sufficient number of points -- if SO, T have < 4 
%	points then an error message is returned.
%
%
%	if either of SO, and T is a scalar then N and M should be used 
%	to specify the range of in 0:max(2*SO,3*X), 0:T. and SIG should of the
%	vector size of same dimension. 
%
%	If both SO, T are scalars the computation is defaulted to
%	blspricefd, SIG has be to scalar in this case.
%
%       if SIG, SO are vectors and T is a scalar then SIG = f(SO) and
%	SIG, SO must be  of same size. if SIG, T are vectors and SO is a scalar
%       then SIG=f(T) and SIG, T must be  of same size. 
%       if SO, T is are both vectors and SIG is a matrix -> SIG = f(SO,T)
%
%	If SIG is a scalar then computation is defaulted to blspricefd.
%
%	You can input SIG as a function of S and T, if SIG is a string
%	then it is assumed to be a MATLAB function [SIG,FofT] = func(S,T)
%	which is expected to work for S, T being both vector arguments.
%	FofT is 1, if SIG is a function of T, else 0 if just a function
%	of stock price.
%
%
%Examples
%========
% 1) SIG =f(SO)
%	Sspace=0:.5:10; Tspace=.1:.1:1;
%	sig=.5*ones(length(Sspace),1)+.1*rand(length(Sspace),1);
%
%	[cfd,pfd]=blspricefdsig(Sspace,5,.1,Tspace,sig);
%
% 2) SIG =f(SO,T)
%	sig=sig=.5*ones(length(Sspace),length(Tspace))+.1*rand(length(Sspace),length(Tspace));
%
%	[cfd,pfd]=blspricefdsig(Sspace,5,.1,Tspace,sig);
%
% 3) : SIG = f(SO) ; f coded in   mysig1.m
%
%	[cfd,pfd]=blspricefdsig(Sspace,5,.1,Tspace,'mysig1');
%
% 4) : SIG = f(SO,T) ; f coded in   mysig.m
%
%	[cfd,pfd]=blspricefdsig(Sspace,5,.1,Tspace,'mysig');
%
%	Also see mysig and mysig1, BLSPRICEFDAMSIG for american options



% Set up default parameters if not provided 
if nargin < 6 q =[]; end
if nargin < 7 M =[]; end
if nargin < 8 N =[]; end
if isempty(q) q=0; end
if isempty(M) M=20; end
if isempty(N) N=20; end

functionofT=0;

call=[];
put=[];
if length(q)~=1 | length(X) ~=1 | length(r)~=1
	error('X, r, q must be scalars');
end

if length(SO)==1 & length(T)==1
if ~ischar(sig) 
        if  length(sig)~=1
        error('sig must be a scalar');
        end
else
        sig=feval(sig,SO,T);
end
		if nargout==1 
			call = blspricefd(SO,X,r,T,sig,q,M,N);
		else
			[call,put] = blspricefd(SO,X,r,T,sig,q,M,N);
		end
else

if length(T)==1
Tvec=linspace(0,T,M);
else
Tvec=T;
end
if length(SO)==1
Svec=linspace(0,max(2*SO,3*X),N);
else
if SO(1)~=0 error('The first element in stock space SO must be 0, SO(1)=0'); end
Svec=SO;
end

Svec=Svec(:);
Tvec=Tvec(:);
M=length(Tvec);
N=length(Svec);

if length(Svec) <=4 | length(Tvec) <=4
	error('Please choose more than 4 points for Stock and time space');
end
if ~ischar(sig) 
        [nr,nc]=size(sig);
	if nr==1 & nc==1
	% Scalar problem
		if nargout==1 
			call = blspricefd(SO,X,r,T,sig,q,M,N);
		else
			[call,put] = blspricefd(SO,X,r,T,sig,q,M,N);
		end
	elseif (nr==1) | (nc==1)
	% sigma vector f(SO)
		sigval = sig(:);
		if length(SO)==1
		if length(sigval)~=M
	error('sigval should be of same dimension as SO x T space - if one of SO or T is entered as a scalar then the dimension is specified by M and N parameters');
		end
		sigval=sigval';
		sigval=repmat(sigval,N,1);
		else
		if length(sigval)~=N
	error('sigval should be of same dimension as SO x T space - if one of SO or T is entered as a scalar then the dimension is specified by M and N parameters');
		end


		end
	else
	% Sig  matrix  = f(SO,T)
		sigval=sig;
		functionofT=1;
		if size(sig,1)~=N | size(sig,2)~=M
	error('sig should be of same dimension as SO x T space - if one of SO or T is entered as a scalar then the dimension is specified by M and N parameters');
		end
	end
else
%	sig = function
	[sigval,functionofT]=feval(sig,Svec,Tvec);
end

if isempty(call) | isempty(put)

% The time step
dt = max(max(Tvec))/(M-1);


% COMPUTE the CALL value.
u=max(Svec-X,0);
count=0;
for i = dt : dt : max(max(Tvec))
        count=count+1;
        if (functionofT) | (count==1)
	% Setup the finite-difference matrix for timestepping.
        [A,c,a,A1,c1,a1]=compA(sigval(:,count),r,q,max(max(Tvec)),M,N);
	end
	u1=timestepcall(u(:,count),A,c,A1,c1,Svec,X,exp(-q*i));
        u=[u u1];
end

call = u;

% COMPUTE the PUT value.
if nargout > 1
u=max(X-Svec,0);

count=0;
for i = dt : dt : max(max(Tvec))
        count=count+1;
        if (functionofT) | (count==1)
	% Setup the finite-difference matrix for timestepping.
        [A,c,a,A1,c1,a1]=compA(sigval(:,count),r,q,max(max(Tvec)),M,N);
	end
	u1=timestepput(u(:,count),A,a,A1,a1,i*r,X);
        u=[u u1];
end
put =u;

end

end
end
