function u = blspricefd(SO,X,r,T,sig,q,M,N)
%
%BLSPRICEFD Black-Scholes put and call pricing for European Options using 
%	the Crank-Nicholson finite difference solution of Black-Scholes
%	Partial differential equation. Note that this function
%	returns an approximate solution unlike the analytical solution (BLSPRICE)
%
%       [CALL,PUT] = BLSPRICEFDAM(SO,X,R,T,SIG,Q,M,N) returns the value of
%       call and put options using the Black-Scholes pricing formula.  SO is
%       the current asset price, X is the exercise price, R is the risk-free
%       interest rate, T is the time to maturity of the option in years, SIG
%       is the standard deviation of the annualized continuously compounded
%       rate of return of the asset (also known as volatility), and Q is the
%       dividend rate of the asset.  The default Q is 0. N denotes the number
%       of discretization points in the stock price domain, and M denotes 
%	the number of discretization points in time domain used for the PDE solution.
%       Try increasing either of M or N to achieve greater efficiency. 
%       
%       For example, the current price an asset is $100, the exercise price 
%       of the option is $95, the risk-free interest rate is 10%, the time 
%       to maturity of the option is .25 years, and the standard deviation of 
%       of the asset is 50%.  Using this data,  
% 
%       [call,put] = blspricefdam(100,95,.1,.25,.5,0) 
% 
%       returns a call option price of $13.64 and put option value of $6.41
%        
%       Reference: John C Hull : Options, Futures and other derivatives, Chap 15.
% 
%       See also BLSPRICEFDAM for American options


% Set up default parameters if not provided 
if nargin < 6 q =[]; end
if nargin < 7 M =[]; end
if nargin < 8 N =[]; end
if isempty(q) q=0; end
if isempty(M) M=20; end
if isempty(N) N=20; end

x1=0;
x2=200;
S=linspace(x1,x2,N); S=S(:);
%S=deriv(S,ones(N,1));
pad = ones(N,1);
newsig=sig(1,1).*pad;

% Setup the finite-difference matrix for timestepping.
[A,c,a,A1,c1,a1]=compA(newsig,r(1,1),q(1,1),max(max(T)),M,N);

% The time step
dt = max(max(T))/(M-1);



% COMPUTE the CALL value.
u=(S>X(1,1));
if max(max(T))==0 tvec=0; else tvec=0:dt:max(max(T)); end
count=0;
for i = dt : dt : max(max(T))
        count=count+1;
	u1=digcallstep(u,A,c,A1,c1,S,X(1,1),exp(-(r-q)*i));
        u=u1;
end

