%-----------------------------------------------------------------------
% Steady-state heat conduction in a composite rod
%
%    2 groups of elements with coincident nodes
%    inputs set up to allow for automatic differentiation
%
%    This is restricted to conforming elements of arbitrary order
%
% Example usage:
%    rod4(1,4,1,0,0,1,0.5,0,0);    <- parameters for the Reno 00 paper
%
%  for a contact resistance, we can use
%    rod4(1,4,1,0,0,1,0.5,0,0,20); <- parameters for the Reno 00 paper
%-----------------------------------------------------------------------
function u = rod4(k1,k2,u1,u2,x1,x2,xi,f1,f2,h)

  maxiter    = 1;
  tol_u      = 1.d-10;
  tol_res    = 1.d-10;
  
  if (nargin == 9)
    penalty    = 10^6;
  else
    penalty    = h;
  end

  if (nargin == 10)
    if (h == 0)
      coincident = 0;
    end
  else
    coincident = 1;
  end

  ngroup     = 2;
  nelgrp(1)  = 10;
  nelgrp(2)  = 10;

  order      = 1;

  npgaus     = 2;  % number of Gauss points for integration

  conduc(1)  = k1;
  conduc(2)  = k2;
  force (2)  = f2;
  
  %---------------------------------------------------------------------
  % generate the mesh (node locations and element connectivity)
  %   this is specifically written for two groups of elements
  %---------------------------------------------------------------------
  ndp    = order + 1;   % number of nodes per element
  nelem  = sum(nelgrp);
  
  xleft  = linspace( x1, xi, nelgrp(1)*order+1 );
  xright = linspace( xi, x2, nelgrp(2)*order+1 );

  if (coincident)
    numnp  = order*nelem + ngroup;

    x      = [ xleft xright ];
  else
    numnp  = order*nelem + 1;
    
    x = [ xleft(1:nelgrp(1)*order) xright ];
  end
    
  nd      = 0;
  element = 0;
  for group=1:ngroup
    for ne = 1:nelgrp(group)
      element = element + 1;
      for i=1:ndp
        nodes(element,i) = nd + i;
      end
      nd = nd + ndp-1;
    end
    nd = nd + 1;
  end

  if (coincident == 1)
    n1 = nodes(nelgrp(1)  ,ndp);
    n2 = nodes(nelgrp(1)+1,  1);
  end

  %---------------------------------------------------------------------
  % initial solution (we are treating this linear problem like a
  % nonlinear one, presumably to eventually consider nonlinear heat
  % conduction problems)
  %---------------------------------------------------------------------
% JEFF These are the statements you need
% Another way for the trick is to also do npd=ndp+0*k if k is independent say

  numnp=deriv(numnp);
  ndp=deriv(ndp);
  nelem=deriv(nelem);
  npgaus=deriv(npgaus);



u = zeros(numnp,1);

  %---------------------------------------------------------------------
  % solution du probleme
  %---------------------------------------------------------------------
  iter      = 1;
  converged = 0;
  diverged  = 0;
  
  while (converged == 0 & diverged == 0)
    % build residual of nonlinear equations
    % -------------------------------------
    A = zeros(numnp,numnp);
    b = zeros(numnp,1);

    % loop on groups of elements
    % --------------------------
    first_e = 0;
    for group = 1:ngroup
      % extract connectivity
      % --------------------
      nodes_gr = zeros(nelgrp(group),ndp);
      for ie = 1:nelgrp(group)
        nodes_gr(ie,1:ndp) = nodes(first_e+ie,1:ndp);
      end
      first_e = first_e + nelgrp(group);
      cond    = conduc(group);
      f       = force (group);
      % loop on elements in group
      % -------------------------
      for ie = 1:nelgrp(group);
        matlem = zeros(ndp,ndp);
        belm   = zeros(ndp,1);
        xe     = zeros(ndp,1);
        ue     = zeros(ndp,1);
        node   = zeros(ndp,1);

        node(1:ndp) = nodes_gr(ie,1:ndp);
        xe  (1:ndp) = x    (node(1:ndp));
        ue  (1:ndp) = u    (node(1:ndp));

        % Gaussian quadrature and shape functions
        % ---------------------------------------
        [xg,wg] = gauss_1d(npgaus);
        [djac, phi,dphidx] = shape_1d(xg,xe);

        % element system via numerical jacobian
        % -------------------------------------
        uk    =    phi'*ue;
        dudxk = dphidx'*ue;

        % element RHS = weak residual
        % ---------------------------
        weak_res = w_residual(cond,phi,dphidx,uk,dudxk,xg,wg,djac,f);
        
        belm =-weak_res;
        
        % Element Matrix = numerical jacobian
        % -----------------------------------
        for nj = 1:ndp
          eps_u  = 1.e-4;
          ue(nj) = ue(nj) + eps_u;
     
          % solution at gauss points:
          % ------------------------
          uk    =    phi'*ue;
          dudxk = dphidx'*ue;

          % Element matrix
          % --------------
          weak_res_plus = w_residual(cond,phi,dphidx,uk,dudxk,xg,wg,djac,f);

          for ni = 1:ndp
            matelm(ni,nj) = (weak_res_plus(ni) - weak_res(ni))/eps_u;
          end
          ue(nj) = ue(nj) - eps_u;
        end
 
        % Assembly
        % --------
        A(node(1:ndp),node(1:ndp)) = A(node(1:ndp),node(1:ndp))...
                                    + matelm(1:ndp,1:ndp);
        b(node(1:ndp))             = b(node(1:ndp))...
                                    + belm(1:ndp);
      end % end loop on elements in group
    end % end loop on groups of elements

    if (coincident==1)
      % Add the constraint on the elements (phi is continuous)
      % via the penalty method
      %-----------------------------------
      matelm = zeros(2,2); belm = zeros(2,1);
      node   = zeros(2,1); ue   = zeros(2,1);
      
      matelm(1,1) = 1*penalty;    matelm(1,2) =-1*penalty;
      matelm(2,1) =-1*penalty;    matelm(2,2) = 1*penalty;
      belm(1)     = 0;
      belm(2)     = 0;

      % put in correction-residual form
      %    to match numerical Jacobian form
      % -----------------------------------
      node(1)     = n1;
      node(2)     = n2; 
      ue(1)       = u(n1);
      ue(2)       = u(n2);

      belm = belm - matelm*ue;
    
      A(node(1:2),node(1:2)) = A(node(1:2),node(1:2)) + matelm(1:2,1:2);
      b(node(1:2))           = b(node(1:2))           + belm(1:2); 
    end
    % Dirichlet conditions
    % --------------------
    A(1    ,1:numnp) = zeros(1    , numnp);
    A(1    ,1      ) = 1;
    b(1)             = u1 - u(1);
    A(numnp,1:numnp) = zeros(1    , numnp);
    A(numnp,numnp  ) = 1;
    b(numnp)         = u2 - u(numnp);
 
    norm_res = norm(b,2);
    [L,U] = lu(A);
    y  = L\b;
    du = U\y;
    u  = u + du;
    norm_du  = norm(du,2);
    norm_u   = norm(u ,2);
    err_u    = norm_du/norm_u;

    disp(sprintf('\nEnd of Iteration %2i',iter))
    disp(sprintf('  norm_res = %12.6e , err_u = %12.6e',norm_res,err_u))
    
    if err_u <= tol_u & norm_res <= tol_res
      [message,erreur] = sprintf('\nsolution convergee en %d iterations',iter);
      disp(message)
      converged = 1;
    end

    if iter > maxiter
      [message,erreur] = sprintf('\nsolution non convergee en %d iterations',iter);
      disp(message)
      diverged = 1;
    end

    iter = iter + 1;
  end % fin des iterations

%  figure(1)
%  plot_solution(x,nelgrp,order,u)
%  grid
%  title('Temperature Profile')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to TL
  %---------------------------------------------------------------------
  b    = zeros(numnp,1);
  b(1) = 1;

  y  = L\b;
  S  = U\y;
%  figure(2)
%  plot_solution(x,nelgrp,order,S)
%  grid
%  title('Sensitivity to Left Boundary Condition')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to Tr
  %---------------------------------------------------------------------
  b        = zeros(numnp,1);
  b(numnp) = 1;

  y  = L\b;
  S  = U\y;

%  figure(3)
%  plot_solution(x,nelgrp,order,S)
%  grid
%  title('Sensitivity to Right Boundary Condition')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to k1
  %---------------------------------------------------------------------
  dkda(1) = 1;  % a = k1
  dkda(2) = 0;
nelem=getval(nelemp);
numnp=getval(numnp);
ndp=getval(ndp);
npgaus=getval(npgaus);
  
  b       = zeros(numnp,1);
  first_e = 0;
  for group = 1:ngroup
    % extract connectivity
    % --------------------
    nodes_gr = zeros(nelgrp(group),ndp);
    for ie = 1:nelgrp(group)
      nodes_gr(ie,1:ndp) = nodes(first_e+ie,1:ndp);
    end
    first_e = first_e + nelgrp(group);
    % loop on elements in group
    % -------------------------
    for ie = 1:nelgrp(group);
      belm = zeros(ndp,1);
      node = zeros(ndp,1);
      xe   = zeros(ndp,1);
      ue   = zeros(ndp,1);

      node(1:ndp) = nodes_gr(ie,1:ndp);
      xe  (1:ndp) = x    (node(1:ndp));
      ue  (1:ndp) = u    (node(1:ndp));

      % Gaussian quadrature and shape functions
      % ---------------------------------------
      [xg,wg] = gauss_1d(npgaus);
      [djac, phi,dphidx] = shape_1d(xg,xe);

      % evaluate the right hand side
      % -------------------------------------
      uk    =    phi'*ue;
      dudxk = dphidx'*ue;

      for i=1:ndp
        for k=1:npgaus
          belm(i) = belm(i) - dkda(group)*dudxk(k)*dphidx(i,k)*wg(k)*djac(k);
        end
      end
      b(node(1:ndp)) = b(node(1:ndp)) + belm(1:ndp);
    end
  end

  b(1)     = 0;
  b(numnp) = 0;

  y  = L\b;
  S  = U\y;

  figure(4)
  plot_solution(x,nelgrp,order,S)
  grid
  title('Sensitivity to Conductivity, k_1')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to k2
  %---------------------------------------------------------------------
  dkda(1) = 0;
  dkda(2) = 1;  % a = k2
  
  b       = zeros(numnp,1);
  first_e = 0;
  for group = 1:ngroup
    % extract connectivity
    % --------------------
    nodes_gr = zeros(nelgrp(group),ndp);
    for ie = 1:nelgrp(group)
      for nd=1:ndp
        nodes_gr(ie,nd) = nodes(first_e+ie,nd);
      end
    end
    first_e = first_e + nelgrp(group);
    % loop on elements in group
    % -------------------------
    for ie = 1:nelgrp(group);
      belm = zeros(ndp,1);
      node = zeros(ndp,1);
      xe   = zeros(ndp,1);
      ue   = zeros(ndp,1);

      node(1:ndp) = nodes_gr(ie,1:ndp);
      xe  (1:ndp) = x    (node(1:ndp));
      ue  (1:ndp) = u    (node(1:ndp));

      % Gaussian quadrature and shape functions
      % ---------------------------------------
      [xg,wg] = gauss_1d(npgaus);
      [djac, phi,dphidx] = shape_1d(xg,xe);

      % evaluate the right hand side
      % -------------------------------------
      uk    =    phi'*ue;
      dudxk = dphidx'*ue;

      for i=1:ndp
        for k=1:npgaus
          belm(i) = belm(i) - dkda(group)*dudxk(k)*dphidx(i,k)*wg(k)*djac(k);
        end
      end
      b(node(1:ndp)) = b(node(1:ndp)) + belm(1:ndp);
    end
  end

  b(1)     = 0;
  b(numnp) = 0;

  y  = L\b;
  S  = U\y;

%  figure(5)
%  plot_solution(x,nelgrp,order,S)
%  grid
%  title('Sensitivity to Conductivity, k_2')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to x1
  %---------------------------------------------------------------------
  b     = zeros(numnp,1);

  ie          = 1;
  node(1:ndp) = nodes(ie,1:ndp);
  xe  (1:ndp) = x (node(1:ndp));
  ue  (1:ndp) = u (node(1:ndp));

  [djac, phi,dphidx] = shape_1d([-1;0;1],xe);
  
  dudx1 = 0;
  for i=1:ndp
    dudx1 = dudx1 + dphidx(i,1)*ue(i);
  end

  b(1)     =-dudx1;
  b(numnp) = 0;
  
  y = L\b;
  S = U\y;
  
%  figure(6)
% plot_solution(x,nelgrp,order,S)
%  grid
%  title('Sensitivity to Left End Point, x_1')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to x2
  %---------------------------------------------------------------------
  b     = zeros(numnp,1);

  ie          = nelem;
  node(1:ndp) = nodes(ie,1:ndp);
  xe  (1:ndp) = x (node(1:ndp));
  ue  (1:ndp) = u (node(1:ndp));

  [djac, phi,dphidx] = shape_1d([-1;0;1],xe);
  
  dudx2 = 0;
  for i=1:ndp
    dudx2 = dudx2 + dphidx(i,3)*ue(i);
  end
  b(1)     = 0;
  b(numnp) =-dudx2;
  
  y = L\b;
  S = U\y;
  
%  figure(7)
%  plot_solution(x,nelgrp,order,S)
%  grid
%  title('Sensitivity to Right End Point, x_2')

  %---------------------------------------------------------------------
  % Compute sensitivity with respect to the interface location, xi
  %---------------------------------------------------------------------
  b           = zeros(numnp,1);
  S           = zeros(numnp,1);
  
  ie          = nelgrp(1);
  node(1:ndp) = nodes(ie,1:ndp);
  xe  (1:ndp) = x (node(1:ndp));
  ue  (1:ndp) = u (node(1:ndp));

  [djac, phi,dphidx] = shape_1d([-1;1],xe);
  
  dudxl = 0;
  for i=1:ndp
    dudxl = dudxl + dphidx(i,2)*ue(i);
  end
  
  ie          = nelgrp(1)+1;
  node(1:ndp) = nodes(ie,1:ndp);
  xe  (1:ndp) = x (node(1:ndp));
  ue  (1:ndp) = u (node(1:ndp));

  [djac, phi,dphidx] = shape_1d([-1;0;1],xe);
  
  dudxr = 0;
  for i=1:ndp
    dudxr = dudxr + dphidx(i,1)*ue(i);
  end

  b(n1) =-penalty*( dudxl-dudxr -S(n1)+S(n2) );
  b(n2) =-penalty*( dudxr-dudxl +S(n1)-S(n2) );
  
  y = L\b;
  S = U\y;
  
%  figure(8)
%  plot_solution(x,nelgrp,order,S)
%  grid
%  title('Sensitivity to Interface Location, x_i')

  %---------------------------------------------------------------------
  % Hack solution to the sensitivity of T with respect to xi
  %---------------------------------------------------------------------
  %  First, compute sensitivity to moving the endpoints ( <-- )
%   b     = zeros(numnp,1);
% 
%   ie          = 1;
%   node(1:ndp) = nodes(ie,1:ndp);
%   xe  (1:ndp) = x (node(1:ndp));
%   ue  (1:ndp) = u (node(1:ndp));
% 
%   [djac, phi,dphidx] = shape_1d([-1;1],xe);
%   
%   dudx1 = 0;
%   for i=1:ndp
%     dudx1 = dudx1 + dphidx(i,1)*ue(i);
%   end
% 
%   b(1)     = dudx1;
%   
%   ie          = nelem;
%   node(1:ndp) = nodes(ie,1:ndp);
%   xe  (1:ndp) = x (node(1:ndp));
%   ue  (1:ndp) = u (node(1:ndp));
% 
%   [djac, phi,dphidx] = shape_1d([-1;0;1],xe);
%   
%   dudx2 = 0;
%   for i=1:ndp
%     dudx2 = dudx2 + dphidx(i,3)*ue(i);
%   end
%   b(numnp) = dudx2;
%   
%   y = L\b;
%   S = U\y;
% 
%   %  Now, add in the derivative of the temperature...
%   for ie=1:nelem
%     node(1:ndp) = nodes(ie,1:ndp);
%     xe  (1:ndp) = x (node(1:ndp));
%     ue  (1:ndp) = u (node(1:ndp));
%      
%     [djac, phi,dphidx] = shape_1d([-1;1],xe);
%      
%     dudx1 = 0;
%     dudx2 = 0;
%     for i=1:ndp
%       dudx1 = dudx1 + dphidx(i,1)*ue(i);
%       dudx2 = dudx2 + dphidx(i,ndp)*ue(i);
%     end
% 
%     dudx(node(1)) = dudx1;
%     dudx(node(2)) = dudx2;
%   end
%   S = S - dudx';
%   
%   figure(9)
%   plot_solution(x,nelgrp,order,S)
%   grid
%   title('Alternate Sensitivity to Interface Location, x_i')

%-----------------------------------------------------------------------
% Function to compute the weak residual of the PDE
%-----------------------------------------------------------------------
function weak_res = w_residual(cond,phi,dphidx,uk,dudxk,xg,wg,djac,f)
  [ndp,npgaus] = size(phi);
  % element RHS = weak residual
  % ---------------------------
  ndp=deriv(ndp);
weak_res = zeros(ndp,1);
  for i = 1:ndp
    for k = 1:npgaus
      weak_res(i) = weak_res(i)...
                  + cond.*dudxk(k).*dphidx(i,k).*wg(k).*djac(k) ...
                  + f.*phi(i,k).*wg(k).*djac(k);
    end
  end

%-----------------------------------------------------------------------
% Function to return Gauss integration points on (-1,1)
%-----------------------------------------------------------------------
function [rg,wg] = gauss_1d(npgaus)

  rg = zeros(npgaus,1);
  wg = zeros(npgaus,1);

  if npgaus == 1
    rg(1) = 0;
    wg(1) = 1;
  elseif npgaus == 2
    rg(1) =-1.0d0 / sqrt(3.0d0);                                  
    rg(2) =-rg(1);
    wg(1) = 1.0;
    wg(2) = 1.0;
  elseif npgaus == 3
    rg(1) =-sqrt(3.0d0/5.0d0);
    rg(2) = 0.0;
    rg(3) =-rg(1);
    wg(1) = 5.0d0 / 9.0d0;
    wg(2) = 8.0d0 / 9.0d0;
    wg(3) = wg(1);
  elseif npgaus == 4
    rg(1) =-sqrt((3.0d0+2.0*sqrt(6.0d0/5.0d0))/7.0d0);
    rg(2) =-sqrt((3.0d0-2.0*sqrt(6.0d0/5.0d0))/7.0d0);
    rg(3) =-rg(2);              
    rg(4) =-rg(1);
    wg(1) = 0.5d0 - 1.0d0 / ( 6.0d0 * sqrt(6.0d0/5.0d0) );
    wg(2) = 0.5d0 + 1.0d0 / ( 6.0d0 * sqrt(6.0d0/5.0d0) );
    wg(3) = wg(2);
    wg(4) = wg(1);
  elseif npgaus == 5
    rg(1) =-sqrt(5.0d0+4.0d0*sqrt(5.0d0/14.0d0)) / 3.0d0;
    rg(2) =-sqrt(5.0d0-4.0d0*sqrt(5.0d0/14.0d0)) / 3.0d0;
    rg(3) = 0.0d0;
    rg(4) =-rg(2);
    rg(5) =-rg(1);
    wg(1) = 161.0d0/450.0d0-13.0d0/(180.d0*sqrt(5.0d0/14.0d0));
    wg(2) = 161.0d0/450.0d0+13.0d0/(180.d0*sqrt(5.0d0/14.0d0));
    wg(3) = 128.0d0/225.0d0;
    wg(4) = wg(2);
    wg(5) = wg(1);
  elseif npgaus == 6
    rg(1) = -0.2386191861;
    rg(2) = -0.6612093865;
    rg(3) = -0.9324695142;
    rg(4) = - rg(1);
    rg(5) = - rg(2);
    rg(6) = - rg(3);
    wg(1) = .4679139346;
    wg(2) = .3607615730;
    wg(3) = .1713244924;
    wg(4) = wg(1);
    wg(5) = wg(2);
    wg(6) = wg(3);
  else
    error('regle de quadrature inconnue')
  end

%-----------------------------------------------------------------------
% Function to plot solution over multiple element groups
%-----------------------------------------------------------------------
function plot_solution(x,nelgrp,order,u)

  ngrp = max(size(nelgrp));

  plot(x(1:nelgrp(1)*order+1),u(1:nelgrp(1)*order+1),'k')
  hold on
  for n=2:ngrp
    n1 = sum(nelgrp(1:n-1))*order+1; n2 = sum(nelgrp(1:n))*order+n;
    plot([x(n1) x(n1+1)], [u(n1) u(n1+1)],':k')
    plot(x(n1+1:n2),u(n1+1:n2),'k')
  end
  hold off


%-----------------------------------------------------------------------
% Function to return shape function and derivatives at Gauss points
%   Currently set up for conforming elements...
%-----------------------------------------------------------------------
function [djac, phi,dphidx] = shape_1d (rg,xe)

  [npgaus,dim] = size(rg);
  [ndp   ,dim] = size(xe);

  order = ndp - 1;
  % Fonctions d'interpolation
  % -------------------------
  if order==1,
    for k = 1:npgaus
      phir  (1,k) = (1-rg(k))/2;
      phir  (2,k) = (1+rg(k))/2;
      dphidr(1,k) =-.5;
      dphidr(2,k) = .5;
    end

  elseif order==2,
    for k=1:npgaus
      phir(1,k) = 0.5*rg(k)*( rg(k)-1 );
      phir(2,k) =-( rg(k)+1 )*( rg(k)-1 );
      phir(3,k) = 0.5*rg(k)*( rg(k)+1 );
    
      dphidr(1,k) = rg(k) - 0.5;
      dphidr(2,k) =-2*rg(k);
      dphidr(3,k) = rg(k) + 0.5;
    
      dphidr2(1,k) = 1;
      dphidr2(2,k) =-2;
      dphidr2(3,k) = 1;
    end    

  elseif order==3,
    for k=1:npgaus
      r2 = rg(k)*rg(k);
      r3 = r2*rg(k);
      
      phir(1,k) =- 9*( r3-r2-rg(k)/9+1/9 )/16;
      phir(2,k) = 27*( r3-r2/3-rg(k)+1/3 )/16;
      phir(3,k) =-27*( r3+r2/3-rg(k)-1/3 )/16;
      phir(4,k) =  9*( r3+r2-rg(k)/9-1/9 )/16;
      
      dphidr(1,k) =- 9*( 3*r2-2*rg(k)-1/9 )/16;
      dphidr(2,k) = 27*( 3*r2-2*rg(k)/3-1 )/16;
      dphidr(3,k) =-27*( 3*r2+2*rg(k)/3-1 )/16;
      dphidr(4,k) =  9*( 3*r2+2*rg(k)-1/9 )/16;
      
      dphidr2(1,k) =- 9*( 6*rg(k)-2   )/16;
      dphidr2(2,k) = 27*( 6*rg(k)-2/3 )/16;
      dphidr2(3,k) =-27*( 6*rg(k)+2/3 )/16;
      dphidr2(4,k) =  9*( 6*rg(k)+2   )/16;
    end

  else
    msg = sprintf('Elements of order %1i are not supported',order);
    error(msg)
  end

  % jacobien de la transformation
  % -----------------------------
  dxdr = dphidr'*xe;
  djac = dxdr;
  drdx = 1./djac;

  % fonctions d'interpolation et derivee globales 
  % ---------------------------------------------
  phi = phir;
  for k = 1:npgaus
    dphidx(:,k) = dphidr(:,k)*drdx(k);
  end
